--[[

A script that parses a file with C function prototypes, creates a
wrapper function for the C function, and redefines the C functions by
using macros.

The purpose with this script is to "Barracudify" existing Lua C
bindings by creating a wrapper function that releases the Barracuda
dispatcher mutex when calling the auxiliary C function.

The parser cannot directly parse a C header file. A file must be
created that contains the C function prototypes only. You may create
this file manually or use the script fext.lua to produce this file.

See usage below:

--]]

local fmt=string.format

-- remove trailing and leading whitespace from string.
local function trim(s)
  return (s:gsub("^%s*(.-)%s*$", "%1"))
end

if #arg ~= 3 then
   print("Usage lua mutex.lua in-h-file out-h-and-c-file main-incfile")
   print("Example: lua mutex.lua funcs.txt myfile funcs.h")
   os.exit(1)
end

local function fopen(name, action)
   local f=io.open(name, action or "r")
   if not f then
      print("Cannot open ",name)
      os.exit(1)
   end
   return f
end

local fpin=fopen(arg[1])


local hdata = fpin:read"*a"
fpin:close()
local fa = {} -- Function array
-- Parse hdata.
-- Iterate prototypes. assume function prototypes end with ';'
for l in string.gmatch(hdata, "([^;]+)") do
   l=trim(l)
   local ok=false
   local fpin -- Function
   -- Extract return type (r1, r2), method name, and arguments
   local r1,r2,m,a=l:match("([%w_]+)([^%w_]*)([^%(]-)%(([^%)]-)%)")
   if r1 then
      ok=true
      fpin = {r1=r1,r2=trim(r2),m=trim(m) }
      a=trim(a)
      if #a > 0 then fpin.a=a end -- Have args
      if fpin.a then
         ok=false
         fpin.an = {} -- The arguments var name only
         -- Iterate arguments
         for ap in string.gmatch(fpin.a, "([^,]+)") do
            ok=true
            -- Extract variable name
            local an = string.match(trim(ap),"([%w_]+)$")
            if an then
               table.insert(fpin.an,an)
            else
               ok=false
               break
            end
         end
         if #fpin.an==1 and fpin.an[1] == "void" then
            fpin.an=nil
            fpin.a=nil
         end
      end
   end
   if ok then
      table.insert(fa,fpin)
   else
      if #l > 0 then
         print("Error, cannot parse: ",l)
         os.exit(1)
      end
   end
end

local hdata={} -- The produced H output
local function emith(...) table.insert(hdata, fmt(...)) end
local cdata={} -- The produced C output
local function emitc(...) table.insert(cdata, fmt(...)) end
local function emit(...)
   local data=fmt(...)
   table.insert(hdata, data)
   table.insert(cdata, data)
end

-- assemble and return the argument name list
function getAnList(fpin, startToken)
   if fpin.a then
      local t={}
      if startToken then
         table.insert(t, startToken)
      end
      local first=true
      for _,an in ipairs(fpin.an) do
         if first then first=false else table.insert(t,", ") end
         table.insert(t,an)
      end
      return table.concat(t)
   end
   return ""
end

emith[[

/*      Auto generated file.
        Generator: mutex.lua
*/

#ifdef BA_LUA_LIB
#include <HttpServer.h>
#include <ThreadLib.h>
#include <balua.h>
ThreadMutex* getDispMutex(lua_State* L);
#else
#include <luaintf.h>
#endif

]]

emith(fmt('#include "%s"\n',arg[3]))

emitc[[

/*      Auto generated file.
        Generator: mutex.lua
*/
]]
emitc(fmt('#include "%s.h"\n',arg[2]:match"[^\\/]+$"))
emitc[[
#ifdef BA_LUA_LIB
ThreadMutex*
getDispMutex(lua_State* L)
{
   return SoDisp_getMutex(HttpServer_getDispatcher(balua_getparam(L)->server));
}
#endif

]]


-- Emit the wrappers

for _,fpin in ipairs(fa) do
   emitc("#undef %s\n",fpin.m)
   local ret = fpin.r1 ~= "void" or fpin.r2 == "*"
   emit("%s %s %s_L(lua_State* L", fpin.r1,fpin.r2,fpin.m)
   if fpin.a then
      emit(", %s",fpin.a)
   end
   emith");\n"
   emitc(")\n{\n   ")
   if ret then emitc("%s %s ret;\n   ",fpin.r1,fpin.r2) end
   emitc"ThreadMutex* __tm = getDispMutex(L);\n   ThreadMutex_release(__tm);\n   "
   if ret then emitc("ret = ") end
   emitc("%s(%s);\n",fpin.m,getAnList(fpin))
   emitc"   ThreadMutex_set(__tm);\n"
   if ret then emitc("   return ret;\n") end
   emitc"}\n"
   emith("#define %s(%s) %s_L(L%s)\n\n" ,fpin.m,getAnList(fpin),fpin.m,getAnList(fpin, ", "))
end

local fp=fopen(fmt("%s.h",arg[2]),"w")
fp:write(table.concat(hdata))
fp:close()
fp=fopen(fmt("%s.c",arg[2]),"w")
fp:write(table.concat(cdata))
fp:close()
